-- Image Resource Viewer - part of the GRIM (Graphic Resource Identity Mapping) project
--
-- ability to filter Library list by width and height?
-- need category maintenance window
-- do we want to track "linked images" such as the shadow and quality masks for items?
-- -- or should we just dump that into the description of the shadow and masks? (probably makes more sense to just put it in description)
-- -- that linking info will only be available from the database anyway, not likely anyone is going to match up those resources manually, THAT would be a daunting task...

-- disable controls during "attempts" to display unknown images

-- for images in "Danger" range, display a special danger icon
-- for images not in "KnownGood" range, display a question mark

-- in single image mode, add label for notes if the image is in "Library" list
-- -- this may not be practical, not sure how much delay this will add when finding image in Library, right now Library is pretty small, but it could get significantly large

-- need to be able to switch an image from "Danger" to "Unknown" (don't allow switching to "KnownGood" without using the actual "attempt" mechanism)- use a right click popup from the traffic light image
-- need to be able to switch an image from "KnownGood" to "Danger" or "Unknown" - use a right click popup from the traffic light image

-- need to be able to import other peoples ranges...
-- need to be able to back up current data with a timestamp, that way if something goes REALLY wrong, we can revert to the backup (load data from back up and perform normal save to replace the "bad" data). need to store the backup file name in a list of backups to allow recovery and removal of extra backups
-- considering how time consuming the image identification process can currently be, losing any known image info would be really painful
-- -- on load, look for import file
-- -- -- if found, read all the ranges
-- -- -- see if any part of the range conflicts with an existing range (i.e. range listed as KnownGood but the current user has some of that range flagged as "Danger")
-- -- -- -- if any of the range conflicts, throw up a warning and ask user if the range should merge only Unknown or should over write the conflicts or abort import
-- -- -- -- -- if the user doesn't abort, track the response with the portion of the range that conflicted, we have to ask and track the seperate response for every overlap
-- -- -- if no abort after reading all ranges,
-- -- -- -- reprocess list, applying ranges using user's choices if there is a conflict
-- -- -- -- only change the range type to "Imported" if it was previously "Default" (or didn't exist, ie was "Unknown" values)
-- -- do we want to provide a button to import files if the plugin is already loaded? should be harmless
-- -- since the file name has to be a fixed value -- we COULD allow the user to type in a file name, but that might just be begging for more trouble than it's worth
-- -- to import multiple files would be a pain if we had to unload each time

-- need to be able to export our ranges
-- -- only export ranges listed as "C" (for Custom) to cut down on overlaps


import "Turbine.UI";
import "Turbine.UI.Lotro";
import "GaranStuff.IRV";

local colorWhite=Turbine.UI.Color(1,1,1);
local colorBlack=Turbine.UI.Color(0,0,0);
local fontColor=Turbine.UI.Color(.9,.9,.1);
local fontName=Turbine.UI.Lotro.Font.TrajanPro14;

UseSafeBoundaries=true;
SafeBoundariesTested=false;
ImageRanges={};
-- ImageRange elements have the structure {Start,End,Class,Type} where Class is "D" or "S" for Danger/Safe and Type is one of "D","I","C" for Default, Imported and Custom respectively
Library={};
-- Library elements have the structure {ID,Description,X,Y,Category} where X and Y are the resolution and Category is a value from Categories[x][1]
Categories={};
-- Categories are the user defined categories for Library, used for filtering
SubCategories={};
SubCategoryCaption={};
DefaultsLoaded=false;
UpdatesLoaded=false;

function UnloadPlugin()
	SaveData(false);
	-- remove shell commands
	Turbine.Shell.RemoveCommand(irvWindow.shellCmd)
	Turbine.Shell.WriteLine("IRV "..Plugins["IRV"]:GetVersion().." by Garan unloaded");
end

function SaveData(triggerSaveEvent)
	Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_ImageRanges", ImageRanges );
	Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_Library", Library );
	Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_Categories", Categories );
	Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_SubCategories", SubCategories );
	Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_SubCategoryCaption", SubCategoryCaption );
	local Settings={};
	local resourceID=tonumber(irvWindow.resourceID:GetText());
	if resourceID==nil then resourceID="" end;
	Settings[1]={"ResourceID",tonumber(resourceID)};
	Settings[2]={"NumberFormat",irvWindow.numberFormat};
	Settings[3]={"UseSafeBoundaries",UseSafeBoundaries};
	Settings[4]={"SafeBoundariesTested",SafeBoundariesTested};
	Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_Settings", Settings );

	if triggerSaveEvent then
		Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_TestID", irvWindow.TestID,SaveComplete());
	else
		Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_TestID", nil);
	end
end

function LoadData()
	ImageRanges=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_ImageRanges");
	Library=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_Library");
	Categories=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_Categories");
	SubCategories=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_SubCategories");
	SubCategoryCaption=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_SubCategoryCaption");
	local testID=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_TestID");
	if ImageRanges==nil then ImageRanges={} end;
	if Library==nil then Library={} end;
	if Categories==nil then Categories={} end;
	if SubCategories==nil then SubCategories={} end;
	if SubCategoryCaption==nil then SubCategoryCaption={} end;
	if testID~=nil then
		-- we were previously testing an "Unknown" ranged resource and it blew up
		-- lock it out to prevent future blowups
		UpdateRange(testID,2);
	end
end

Window = class( Turbine.UI.Window );

function Window:Constructor()
	Turbine.UI.Window.Constructor( self );
	self.Loading=true;
	self:SetWantsUpdates( true );
	self.ShowGrid=false;
	self.AutoPrev=false;
	self.AutoNext=false;
	self.delayTime=0;
	self.delayTotal=20;
	self.displayingWarning=false;
	self.numberFormat="0x%x";
	self.testingBoundaries=false;
	self.mergingUpdate=false;

	local displayWidth, displayHeight = Turbine.UI.Display.GetSize();
	local windowWidth, windowHeight = Turbine.UI.Display.GetSize();
	local positionLeft=-1; -- default it to an invalid value so we know if we get a saved value
	local positionTop=-1;
	local Settings=Turbine.PluginData.Load( Turbine.DataScope.Account, "IRV_Settings");
	local lastResourceID;
	if Settings ~=nil then
		local lIndex;
		for lIndex=1,#Settings do
			if Settings[lIndex][1]=="Width" then
				if Settings[lIndex][2]<windowWidth then
					windowWidth=Settings[lIndex][2];
				end
			end
			if Settings[lIndex][1]=="Height" then
				if Settings[lIndex][2]<windowHeight then
					windowHeight=Settings[lIndex][2];
				end
			end
			if Settings[lIndex][1]=="Left" then
				if Settings[lIndex][2]<positionLeft then
					positionLeft=Settings[lIndex][2];
				end
			end
			if Settings[lIndex][1]=="Top" then
				if Settings[lIndex][2]<positionTop then
					positionTop=Settings[lIndex][2];
				end
			end
			if Settings[lIndex][1]=="NumberFormat" then
				self.numberFormat=Settings[lIndex][2];
			end
			if Settings[lIndex][1]=="ResourceID" then
				lastResourceID=tonumber(Settings[lIndex][2]);
			end
			if Settings[lIndex][1]=="UseSafeBoundaries" then
				UseSafeBoundaries=Settings[lIndex][2];
			end
			if Settings[lIndex][1]=="SafeBoundariesTested" then
				SafeBoundariesTested=Settings[lIndex][2];
			end
		end
	end
	if lastResourceID==nil then lastResourceID=0x41003000 end;

	self:SetSize(windowWidth,windowHeight);
	if positionLeft==-1 then positionLeft=( displayWidth - windowWidth ) / 2 end;
	if positionTop==-1 then positionTop=( displayWidth - windowWidth ) / 2 end;
	self:SetPosition( positionLeft, positionTop);
	self:SetBackColor(Turbine.UI.Color(0,0,0));
	self:SetOpacity(.95);
	self:SetText("Image Resource Viewer");

-- stuff for testing background ids

	self.resourceCaption=Turbine.UI.Label();
	self.resourceCaption:SetParent(self);
	self.resourceCaption:SetSize(90,18);
	self.resourceCaption:SetPosition(0,0);
	self.resourceCaption:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
	self.resourceCaption:SetText("Resource ID:");

	self.resourceID=Turbine.UI.Lotro.TextBox();
	self.resourceID:SetParent(self);
	self.resourceID:SetPosition(self.resourceCaption:GetLeft()+self.resourceCaption:GetWidth()+5,self.resourceCaption:GetTop()-2)
	self.resourceID:SetSize(100,18);
	self.resourceID:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
	self.resourceID:SetBackColor(Turbine.UI.Color(.1,.1,.5));
	self.resourceID:SetForeColor(colorWhite);
	self.resourceID:SetText(string.format(self.numberFormat,lastResourceID));

	self.resourceDescription=Turbine.UI.Label();
	self.resourceDescription:SetParent(self);
	self.resourceDescription:SetSize(self:GetWidth(),18);
	self.resourceDescription:SetPosition(0,24);
	self.resourceDescription:SetFont(Turbine.UI.Lotro.Font.Verdana14);
	self.resourceDescription:SetText("");

	self.imageSize=Turbine.UI.Label();
	self.imageSize:SetParent(self);
	self.imageSize:SetSize(70,18);
	self.imageSize:SetPosition(self.resourceID:GetLeft()+self.resourceID:GetWidth()+5,self.resourceCaption:GetTop());
	self.imageSize:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);

	self.imageRangeIndicator=Turbine.UI.Label();
	self.imageRangeIndicator:SetParent(self);
	self.imageRangeIndicator:SetSize(9,20);
	self.imageRangeIndicator:SetPosition(self.imageSize:GetLeft()+self.imageSize:GetWidth()+5,self.resourceCaption:GetTop())
	self.imageRangeIndicator.Range=2;
	self.imageRangeIndicator.MouseClick=function()
		if self.imageRangeIndicator.Range==1 then
			self.rangeMenu:GetItems():Get(1):SetEnabled(true);
			self.rangeMenu:GetItems():Get(2):SetEnabled(true);
			self.rangeMenu:GetItems():Get(3):SetEnabled(false);
		elseif self.imageRangeIndicator.Range==2 then
			self.rangeMenu:GetItems():Get(1):SetEnabled(false);
			self.rangeMenu:GetItems():Get(2):SetEnabled(true);
			self.rangeMenu:GetItems():Get(3):SetEnabled(true);
		else
			self.rangeMenu:GetItems():Get(1):SetEnabled(true);
			self.rangeMenu:GetItems():Get(2):SetEnabled(false);
			self.rangeMenu:GetItems():Get(3):SetEnabled(true);
		end
		self.rangeMenu:ShowMenu();
	end

	self.imageSolo=Turbine.UI.Label();
	self.imageSolo:SetParent(self);
	self.imageSolo:SetVisible(false)
	self.imageSolo:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.imageSolo.Status=0;
	self.imageSolo.MouseClick=function(sender,args)
		if args.Button==2 then
			if self.imageSolo.Status==1 then
				self.imageMenu:GetItems():Get(1):SetText("");
				self.imageMenu:GetItems():Get(2):SetText("");
				self.imageMenu:GetItems():Get(3):SetText("");
				self.imageMenu:GetItems():Get(4).Click=function()
					editLibWindow:SetResource(tonumber(self.resourceID:GetText()),nil,self.imageWidth,self.imageHeight);
					editLibWindow:SetVisible(true);
				end
				self.imageMenu:ShowMenu();
			end
		end
	end
	self.imageGrid={};

	self.UnknownImageWarning=Turbine.UI.Label();
	self.UnknownImageWarning:SetParent(self);
	self.UnknownImageWarning:SetSize(400,300);
	self.UnknownImageWarning:SetPosition((self:GetWidth()-400)/2,(self:GetHeight()-300)/2);
	self.UnknownImageWarning:SetFont(Turbine.UI.Lotro.Font.Verdana20);
	self.UnknownImageWarning:SetBackColor(Turbine.UI.Color(.1,.1,.1));
	self.UnknownImageWarning:SetForeColor(Turbine.UI.Color(1,0,0));
	self.UnknownImageWarning:SetTextAlignment( Turbine.UI.ContentAlignment.TopLeft);
	self.UnknownImageWarning:SetText("You are attempting to display a previously unverified image.\n\nThe system is currently saving all data in case the image causes a client crash.  As soon as the current data is saved, the image will be displayed.\n\nIf the client does crash, this image will automatically be locked out from displaying in the future (to avoid additional crashes). If the image loads without incident, the image will be flagged as good for future displays.");
	self.UnknownImageWarning:SetVisible(false);
	self.UnknownImageWarning.status=Turbine.UI.Label();
	self.UnknownImageWarning.status:SetParent(self.UnknownImageWarning);
	self.UnknownImageWarning.status:SetSize(0,8);
	self.UnknownImageWarning.status:SetPosition(10,self.UnknownImageWarning:GetHeight()-34);
	self.UnknownImageWarning.status:SetBackColor(Turbine.UI.Color(1,1,0));
	self.UnknownImageWarning.statusBack=Turbine.UI.Label();
	self.UnknownImageWarning.statusBack:SetParent(self.UnknownImageWarning);
	self.UnknownImageWarning.statusBack:SetSize(self.UnknownImageWarning:GetWidth()-20,10);
	self.UnknownImageWarning.statusBack:SetPosition(10,self.UnknownImageWarning:GetHeight()-35);
	self.UnknownImageWarning.statusBack:SetBackColor(Turbine.UI.Color(1,0,0));
	self.UnknownImageWarning.status:SetZOrder(self.UnknownImageWarning.statusBack:GetZOrder()+1);
	self.UnknownImageWarning.btnSkip=Turbine.UI.Lotro.Button();
	self.UnknownImageWarning.btnSkip:SetParent(self.UnknownImageWarning);
	self.UnknownImageWarning.btnSkip:SetSize(100,20);
	self.UnknownImageWarning.btnSkip:SetPosition(15,self.UnknownImageWarning:GetHeight()-22);
	self.UnknownImageWarning.btnSkip:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
	self.UnknownImageWarning.btnSkip.Click=function(sender,args)
		self.delayTime=Turbine.Engine:GetGameTime();
	end
	self.UnknownImageWarning.btnSkip:SetText("Skip Save");

	self.UnknownImageWarning.btnAbort=Turbine.UI.Lotro.Button();
	self.UnknownImageWarning.btnAbort:SetParent(self.UnknownImageWarning);
	self.UnknownImageWarning.btnAbort:SetSize(100,20);
	self.UnknownImageWarning.btnAbort:SetPosition(self.UnknownImageWarning:GetWidth()-self.UnknownImageWarning.btnAbort:GetWidth()-15,self.UnknownImageWarning:GetHeight()-22);
	self.UnknownImageWarning.btnAbort:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
	self.UnknownImageWarning.btnAbort.Click=function(sender,args)
		self.displayingWarning=false;
		self:SetWantsUpdates(false);
		self.AutoPrev=false;
		self.AutoNext=false;
		self.UnknownImageWarning:SetVisible(false);
		irvWindow.TestID=nil;
		Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_TestID",nil);
		self:EnableControls();
	end
	self.UnknownImageWarning.btnAbort:SetText("Abort");


	local width,height=120,120;
	local gridSpace=6;
	self.GridRows=math.floor(748/(height+gridSpace));
	self.GridCols=math.floor(1024/(width+gridSpace));
	local gridLeft=(windowWidth-width*self.GridCols-gridSpace*(self.GridCols-1))/2
	local gridTop=(windowHeight-20-height*(self.GridRows)-gridSpace*(self.GridRows-1))/2

	for iRow=1, self.GridRows do
		self.imageGrid[iRow]={}
		for iCol=1, self.GridCols do
			self.imageGrid[iRow][iCol]=Turbine.UI.Label();
			self.imageGrid[iRow][iCol].Status=0;
			self.imageGrid[iRow][iCol].imageX=0;
			self.imageGrid[iRow][iCol].imageY=0;
			self.imageGrid[iRow][iCol]:SetParent(self);
			self.imageGrid[iRow][iCol]:SetSize(width,height);
			self.imageGrid[iRow][iCol]:SetPosition(gridLeft+(iCol-1)*(width+gridSpace),gridTop+(iRow-1)*(height+gridSpace));
			self.imageGrid[iRow][iCol]:SetBackColor(Turbine.UI.Color(.1,.1,.1));
			self.imageGrid[iRow][iCol].imageRangeIndicator=Turbine.UI.Label();
			self.imageGrid[iRow][iCol].imageRangeIndicator:SetParent(self.imageGrid[iRow][iCol])
			self.imageGrid[iRow][iCol].imageRangeIndicator:SetSize(9,20);
			self.imageGrid[iRow][iCol].imageRangeIndicator:SetPosition(width-10,7);
			self.imageGrid[iRow][iCol].MouseClick=function()
				self.resourceID:SetText(self.imageGrid[iRow][iCol].imageID:GetText())
				self.ShowGrid=false;
				self.btnGridToggle:SetBackground("GaranStuff/IRV/Resources/ToggleSingle.jpg");
				self:ShowBackgrounds(self,false);
			end

			self.imageGrid[iRow][iCol].image=Turbine.UI.Label();
			self.imageGrid[iRow][iCol].image:SetParent(self.imageGrid[iRow][iCol])
			self.imageGrid[iRow][iCol].image:SetSize(32,32);
			self.imageGrid[iRow][iCol].image:SetPosition((width-32)/2,1);
			self.imageGrid[iRow][iCol].image:SetBlendMode(Turbine.UI.BlendMode.Overlay);
			self.imageGrid[iRow][iCol].Cat="";
			self.imageGrid[iRow][iCol].SubCat="";
			self.imageGrid[iRow][iCol].Desc="";
			self.imageGrid[iRow][iCol].ResourceID=0;
			self.imageGrid[iRow][iCol].image.MouseClick=function(sender,args)
				if args.Button==1 then
					self.resourceID:SetText(self.imageGrid[iRow][iCol].imageID:GetText())
					self.ShowGrid=false;
					self.btnGridToggle:SetBackground("GaranStuff/IRV/Resources/ToggleSingle.jpg");
					self:ShowBackgrounds(self,false);
				else
					if self.imageGrid[iRow][iCol].Status==1 then
						self.imageGrid[iRow][iCol].Cat, self.imageGrid[iRow][iCol].SubCat, self.imageGrid[iRow][iCol].Desc=GetLibraryInfo(self.imageGrid[iRow][iCol].ResourceID);
						self.imageMenu:GetItems():Get(1):SetText(tostring(self.imageGrid[iRow][iCol].Cat));
						self.imageMenu:GetItems():Get(2):SetText(self.imageGrid[iRow][iCol].SubCat);
						self.imageMenu:GetItems():Get(3):SetText(self.imageGrid[iRow][iCol].Desc);
						self.imageMenu:GetItems():Get(4).Click=function()
							editLibWindow:SetResource(tonumber(self.imageGrid[iRow][iCol].imageID:GetText()),nil,self.imageGrid[iRow][iCol].imageX,self.imageGrid[iRow][iCol].imageY);
							editLibWindow:SetVisible(true);
						end
						self.imageMenu:ShowMenu();
					end
				end
			end

			self.imageGrid[iRow][iCol].imageID=Turbine.UI.Label();
			self.imageGrid[iRow][iCol].imageID:SetParent(self.imageGrid[iRow][iCol])
			self.imageGrid[iRow][iCol].imageID:SetSize(width,14);
			self.imageGrid[iRow][iCol].imageID:SetPosition(0,height-28);
			self.imageGrid[iRow][iCol].imageID:SetFont(Turbine.UI.Lotro.Font.TrajanPro12);
			self.imageGrid[iRow][iCol].imageID:SetBackColor(Turbine.UI.Color(.2,.2,.3));
			self.imageGrid[iRow][iCol].imageID:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft);

			self.imageGrid[iRow][iCol].imageSize=Turbine.UI.Label();
			self.imageGrid[iRow][iCol].imageSize:SetParent(self.imageGrid[iRow][iCol])
			self.imageGrid[iRow][iCol].imageSize:SetSize(width,14);
			self.imageGrid[iRow][iCol].imageSize:SetPosition(0,height-14);
			self.imageGrid[iRow][iCol].imageSize:SetFont(Turbine.UI.Lotro.Font.TrajanPro12);
			self.imageGrid[iRow][iCol].imageSize:SetBackColor(Turbine.UI.Color(.2,.2,.3));
			self.imageGrid[iRow][iCol].imageSize:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft);
			self.imageGrid[iRow][iCol]:SetVisible(false);
		end
	end

-- create a button to test prev group of backgrounds
	self.btnPriorRange=Turbine.UI.Button();
	self.btnPriorRange:SetParent(self)
	self.btnPriorRange:SetSize(26,20);
	self.btnPriorRange:SetPosition(self.imageRangeIndicator:GetLeft()+self.imageRangeIndicator:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnPriorRange:SetBackground("GaranStuff/IRV/Resources/PriorRangeButton.jpg")
	self.btnPriorRange.Click=function(sender,args)
		-- get the current range index, find the Prior good range with a lower index
		local range,rangeIndex=GetResourceRange(tonumber(self.resourceID:GetText()))
		rangeIndex=rangeIndex-1;
		while rangeIndex>0 and ImageRanges[rangeIndex][3]=="D" do
			rangeIndex=rangeIndex-1;
		end
		if rangeIndex>0 then
			self.resourceID:SetText(string.format(self.numberFormat,ImageRanges[rangeIndex][2]));
			self:ShowBackgrounds(self,false);
		end
	end

	self.btnPrevIDs=Turbine.UI.Button();
	self.btnPrevIDs:SetParent(self)
	self.btnPrevIDs:SetSize(26,20);
	self.btnPrevIDs:SetPosition(self.btnPriorRange:GetLeft()+self.btnPriorRange:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnPrevIDs:SetBackground("GaranStuff/IRV/Resources/BlankButton.jpg")
	self.btnPrevIDs.Image=Turbine.UI.Label();
	self.btnPrevIDs.Image:SetParent(self.btnPrevIDs);
	self.btnPrevIDs.Image:SetSize(16,16);
	self.btnPrevIDs.Image:SetPosition((self.btnPrevIDs:GetWidth()-16)/2,(self.btnPrevIDs:GetHeight()-16)/2)
	self.btnPrevIDs.Image:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.btnPrevIDs.Image:SetBackground(0x41007e0f);
	self.btnPrevIDs.Image:SetMouseVisible(false);
	self.btnPrevIDs.Click=function(sender,args)
		if self.ShowGrid==true then
			-- decrement self.resourceID
			self.resourceID:SetText(string.format(self.numberFormat,tonumber(self.resourceID:GetText())-self.GridRows*self.GridCols));
		else
			self.resourceID:SetText(string.format(self.numberFormat,tonumber(self.resourceID:GetText())-1));
		end
		self:ShowBackgrounds(self,false);
	end

	self.btnGoToImage=Turbine.UI.Button();
	self.btnGoToImage:SetParent(self)
	self.btnGoToImage:SetSize(26,20);
	self.btnGoToImage:SetPosition(self.btnPrevIDs:GetLeft()+self.btnPrevIDs:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnGoToImage:SetBackground("GaranStuff/IRV/Resources/BlankButton.jpg")
	self.btnGoToImage.Image=Turbine.UI.Label();
	self.btnGoToImage.Image:SetParent(self.btnGoToImage);
	self.btnGoToImage.Image:SetSize(18,18);
	self.btnGoToImage.Image:SetPosition((self.btnGoToImage:GetWidth()-18)/2,(self.btnGoToImage:GetHeight()-18)/2)
	self.btnGoToImage.Image:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.btnGoToImage.Image:SetBackground(0x41007dd4);
	self.btnGoToImage.Image:SetMouseVisible(false);
	self.btnGoToImage.Click=function(sender,args)
		self:ShowBackgrounds(self,false);
	end

	self.btnNextIDs=Turbine.UI.Button();
	self.btnNextIDs:SetParent(self)
	self.btnNextIDs:SetSize(26,20);
	self.btnNextIDs:SetPosition(self.btnGoToImage:GetLeft()+self.btnGoToImage:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnNextIDs:SetBackground("GaranStuff/IRV/Resources/BlankButton.jpg")
	self.btnNextIDs.Image=Turbine.UI.Label();
	self.btnNextIDs.Image:SetParent(self.btnNextIDs);
	self.btnNextIDs.Image:SetSize(16,16);
	self.btnNextIDs.Image:SetPosition((self.btnNextIDs:GetWidth()-16)/2,(self.btnNextIDs:GetHeight()-16)/2)
	self.btnNextIDs.Image:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.btnNextIDs.Image:SetBackground(0x41007e12);
	self.btnNextIDs.Image:SetMouseVisible(false);
	self.btnNextIDs.Click=function(sender,args)
		if self.ShowGrid==true then
			-- increment self.resourceID
			self.resourceID:SetText(string.format(self.numberFormat,tonumber(self.resourceID:GetText())+self.GridRows*self.GridCols));
		else
			self.resourceID:SetText(string.format(self.numberFormat,tonumber(self.resourceID:GetText())+1));
		end
		self:ShowBackgrounds(self,false);
	end

	self.btnNextRange=Turbine.UI.Button();
	self.btnNextRange:SetParent(self)
	self.btnNextRange:SetSize(26,20);
	self.btnNextRange:SetPosition(self.btnNextIDs:GetLeft()+self.btnNextIDs:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnNextRange:SetBackground("GaranStuff/IRV/Resources/NextRangeButton.jpg")
	self.btnNextRange.Click=function(sender,args)
		-- get the current range index, find the next good range with a lower index
		local range,rangeIndex=GetResourceRange(tonumber(self.resourceID:GetText()))
		if range~=0 or ImageRanges[rangeIndex][3]=="D" then
			rangeIndex=rangeIndex+1;
			while rangeIndex<=#ImageRanges and ImageRanges[rangeIndex][3]=="D" do
				rangeIndex=rangeIndex+1;
			end
		end
		if rangeIndex<=#ImageRanges then
			self.resourceID:SetText(string.format(self.numberFormat,ImageRanges[rangeIndex][1]));
			self:ShowBackgrounds(self,false);
		end
	end

	self.btnForceView=Turbine.UI.Lotro.Button();
	self.btnForceView:SetParent(self)
	self.btnForceView:SetSize(75,20);
	self.btnForceView:SetPosition(self.btnNextRange:GetLeft()+self.btnNextRange:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnForceView:SetFont(Turbine.UI.Lotro.Font.TrajanPro14);
	self.btnForceView:SetText("Attempt");
	self.btnForceView:SetVisible(false);
	self.btnForceView.Click=function(sender,args)
		self:ShowBackgrounds(self,true);
	end

	self.btnGridToggle=Turbine.UI.Button();
	self.btnGridToggle:SetParent(self)
	self.btnGridToggle:SetSize(40,20);
	self.btnGridToggle:SetPosition(self.btnForceView:GetLeft()+self.btnForceView:GetWidth()+5,self.resourceCaption:GetTop());
	self.btnGridToggle:SetBackground("GaranStuff/IRV/Resources/ToggleSingle.jpg");
	self.btnGridToggle.Click=function(sender,args)
		self.ShowGrid= not self.ShowGrid
		if self.ShowGrid==true then
			self.btnGridToggle:SetBackground("GaranStuff/IRV/Resources/ToggleGrid.jpg");
		else
			self.btnGridToggle:SetBackground("GaranStuff/IRV/Resources/ToggleSingle.jpg");
		end
		self:ShowBackgrounds(self,false);
	end

	self.btnFormatToggle=Turbine.UI.Button();
	self.btnFormatToggle:SetParent(self)
	self.btnFormatToggle:SetSize(40,20);
	self.btnFormatToggle:SetPosition(self.btnGridToggle:GetLeft()+self.btnGridToggle:GetWidth()+6,self.resourceCaption:GetTop());
	if self.numberFormat=="0x%x" then
		self.btnFormatToggle:SetBackground("GaranStuff/IRV/Resources/ToggleHex.jpg");
	else
		self.btnFormatToggle:SetBackground("GaranStuff/IRV/Resources/ToggleDec.jpg");
	end
	self.btnFormatToggle.Click=function(sender,args)
		if self.numberFormat=="0x%x" then
			self.numberFormat="%d";
			self.btnFormatToggle:SetBackground("GaranStuff/IRV/Resources/ToggleDec.jpg");
		else
			self.numberFormat="0x%x";
			self.btnFormatToggle:SetBackground("GaranStuff/IRV/Resources/ToggleHex.jpg");
		end
		self.resourceID:SetText(string.format(self.numberFormat,self.resourceID:GetText()));
		self:ShowBackgrounds(self,false);
	end

	self.btnSearch=Turbine.UI.Button();
	self.btnSearch:SetParent(self)
	self.btnSearch:SetSize(26,20);
	self.btnSearch:SetPosition(self.btnFormatToggle:GetLeft()+self.btnFormatToggle:GetWidth()+6,self.resourceCaption:GetTop());
	self.btnSearch:SetBackground("GaranStuff/IRV/Resources/SearchButton.jpg");
	self.btnSearch.Click=function(sender,args)
		libWindow:SetVisible(true);
	end

	self.Update = function(sender,args)
		if not SafeBoundariesTested then
			-- need to override the default text of the UnknownImageWarning, hide the skip and abort buttons, save the test settings and test one of the known boundaries
			self.TestID=nil;
			self:DisableControls(); -- disable the main window controls so the user doesn't affect the test
			self.UnknownImageWarning:SetText("IRV is testing a known \"Range Boundary\" to determine whether your system requires \"Safe Boundaries\".  When the timer expires, either the client will crash in which case you will need to restart the client and reload IRV at which time, Safe Boundaries will automatically be turned on (the test only runs once).  If the test message disappears with no other result, you do not have to use Safe Boundaries and they will remain off, you can use the plugin immediately.");
			self.UnknownImageWarning.btnSkip:SetVisible(false);
			self.UnknownImageWarning.btnAbort:SetVisible(false);
			self.UnknownImageWarning:SetVisible(true);
			UseSafeBoundaries=true;
			SafeBoundariesTested=true;
			SaveData();
			self.delayTime=Turbine.Engine:GetGameTime()+self.delayTotal;
			self.testingBoundaries=true;
			self.UnknownImageWarning.status:SetWidth(0);
		elseif self.testingBoundaries then
			local curTime=Turbine.Engine:GetGameTime();
			local delta=1-(self.delayTime-curTime)/self.delayTotal;
			if delta>0 then
				self.UnknownImageWarning.status:SetWidth((self.UnknownImageWarning:GetWidth()-20)*delta);
			end
			if self.delayTime<Turbine.Engine:GetGameTime() then
				self:SetWantsUpdates(false);
				TestComplete();
			end
		else
			if self.Loading and (Plugins["IRV"] ~= nil) then
				Plugins["IRV"].Unload = function(self,sender,args)
					UnloadPlugin();
				end
				self.Loading=false;
				Turbine.Shell.WriteLine("IRV "..Plugins["IRV"]:GetVersion().." by Garan loaded");
				self:SetWantsUpdates(false);
			elseif self.displayingWarning then
				local curTime=Turbine.Engine:GetGameTime();
				local delta=1-(self.delayTime-curTime)/self.delayTotal;
				if delta>0 then
					self.UnknownImageWarning.status:SetWidth((self.UnknownImageWarning:GetWidth()-20)*delta);
				end

				if self.delayTime<Turbine.Engine:GetGameTime() then
					self:SetWantsUpdates(false);
					SaveComplete2();
				end
			end
		end
	end       

	self.btnAutoPrev=Turbine.UI.Button();
	self.btnAutoPrev:SetParent(self)
	self.btnAutoPrev:SetSize(26,20);
	self.btnAutoPrev:SetPosition(self:GetWidth()-214,self.resourceCaption:GetTop());
	self.btnAutoPrev:SetBackground("GaranStuff/IRV/Resources/AutoPrevButton.jpg");
	self.btnAutoPrev.Click=function(sender,args)
		-- first, make sure we're in solo mode
		if self.ShowGrid then
			self.ShowGrid=false;
			self:ShowBackgrounds(self,false);
		end
		-- make sure we're on an image with range=0
		self.resourceID:SetText(string.format(self.numberFormat,self.FindUnknownPrev(tonumber(self.resourceID:GetText()))));
		if tonumber(self.resourceID:GetText())>0 then
			self.AutoPrev=true;
			self.AutoNext=false; -- just in case...
			-- get image. the rest of the 'auto' code is in SaveComplete2()
			self:ShowBackgrounds(self,true);
		end
	end

	self.btnAutoNext=Turbine.UI.Button();
	self.btnAutoNext:SetParent(self)
	self.btnAutoNext:SetSize(26,20);
	self.btnAutoNext:SetPosition(self:GetWidth()-183,self.resourceCaption:GetTop());
	self.btnAutoNext:SetBackground("GaranStuff/IRV/Resources/AutoNextButton.jpg");
	self.btnAutoNext.Click=function(sender,args)
		-- first, make sure we're in solo mode
		if self.ShowGrid then
			self.ShowGrid=false;
			self:ShowBackgrounds(self,false);
		end
		-- make sure we're on an image with range=0
		self.resourceID:SetText(string.format(self.numberFormat,self.FindUnknownNext(tonumber(self.resourceID:GetText()))));
		if tonumber(self.resourceID:GetText())>0 then
			self.AutoNext=true;
			self.AutoPrev=false; -- just in case...
			-- get image. the rest of the 'auto' code is in SaveComplete2()
			self:ShowBackgrounds(self,true);
		end
	end

	self.btnIO=Turbine.UI.Button();
	self.btnIO:SetParent(self)
	self.btnIO:SetSize(26,20);
	self.btnIO:SetPosition(self:GetWidth()-121,self.resourceCaption:GetTop());
	self.btnIO:SetBackground("GaranStuff/IRV/Resources/IOButton.jpg");
	self.btnIO.Click=function(sender,args)
		-- put IO stuff here
		ioWindow:SetVisible(true);
	end

	self.btnClose=Turbine.UI.Button();
	self.btnClose:SetParent(self)
	self.btnClose:SetSize(26,20);
	self.btnClose:SetPosition(self:GetWidth()-28,self.resourceCaption:GetTop());
	self.btnClose:SetBackground("GaranStuff/IRV/Resources/CloseButton.jpg");
	self.btnClose.Click=function(sender,args)
		self:SetVisible(false);
	end

	self.resourceCaption:SetZOrder(self.imageSolo:GetZOrder()+1)
	self.resourceID:SetZOrder(self.resourceCaption:GetZOrder()+1)
	self.resourceDescription:SetZOrder(self.resourceID:GetZOrder()+1)
	self.imageSize:SetZOrder(self.resourceDescription:GetZOrder()+1)
	self.imageRangeIndicator:SetZOrder(self.imageSize:GetZOrder()+1)
	self.btnPrevIDs:SetZOrder(self.imageRangeIndicator:GetZOrder()+1)
	self.btnForceView:SetZOrder(self.btnPrevIDs:GetZOrder()+1)
	self.btnNextIDs:SetZOrder(self.btnForceView:GetZOrder()+1)
	self.btnNextRange:SetZOrder(self.btnNextIDs:GetZOrder()+1)
	self.btnGridToggle:SetZOrder(self.btnNextRange:GetZOrder()+1)
	self.btnFormatToggle:SetZOrder(self.btnGridToggle:GetZOrder()+1)
	self.UnknownImageWarning:SetZOrder(self.imageSolo:GetZOrder()+1);

	-- capture Key events
	self.KeyDown=function(sender, args)
--Turbine.Shell.WriteLine("KeyDown Action="..args.Action..", .Alt="..tostring(args.Alt)..", .Control="..tostring(args.Control)..", .Shift="..tostring(args.Shift))
		if ( args.Action == Turbine.UI.Lotro.Action.Escape ) then
			sender:SetVisible( false );
		end
	end

	self.VisibleChanged=function()
		if (not self:IsVisible()) then
			-- hide all the other windows when we are hidden
			if libWindow~=nil then
				libWindow:SetVisible(false);
			end
			if editLibWindow~=nil then
				editLibWindow:SetVisible(false);
			end
			if catMaintWindow~=nil then
				catMaintWindow:SetVisible(false);
			end
			if ioWindow~=nil then
				ioWindow:SetVisible(false);
			end
		end
	end

	-- register shell commands
	self.shellCmd=Turbine.ShellCommand();
	local numberOfCommandsRegistered = Turbine.Shell.AddCommand("irv;ImageResourceViewer",self.shellCmd);
	self.shellCmd.Execute = function(sender, cmd, args)
		if (cmd == "irv" or cmd=="ImageResourceViewer") then
			if (args=="show") then
				self:SetVisible(true);
			elseif (args=="hide") then
				self:SetVisible(false);
			elseif (args=="toggle") then
				self:SetVisible(not self:IsVisible());
			else
				Turbine.Shell.WriteLine("usage:");
				Turbine.Shell.WriteLine("  /irv show");
				Turbine.Shell.WriteLine("       Displays the Image Resource Viewer.");
				Turbine.Shell.WriteLine("  /irv hide");
				Turbine.Shell.WriteLine("       Hides the Image Resource Viewer.");
				Turbine.Shell.WriteLine("  /irv toggle");
				Turbine.Shell.WriteLine("       Toggles the Image Resource Viewer (shows it if it was hidden, hides it if it was showing).");
			end
		end
	end
	self.shellCmd.GetHelp = function(sender, cmd)
		return("usage:\n  /irv show\n       Displays the Image Resource Viewer.\n  /irv hide\n       Hides the Image Resource Viewer.\n  /irv toggle\n       Toggles the Image Resource Viewer (shows it if it was hidden, hides it if it was showing).");
	end

	self.shellCmd.GetShortHelp = function(sender, cmd)
		return("irv - usage: /irv show | hide | toggle");
	end
	self:SetWantsKeyEvents(true);

	self.FindUnknownPrev=function(resourceID)
		-- get the resourcerange of the resourceID
		local range,rangeID=GetResourceRange(resourceID);
		while resourceID>0 and range>0 do
			resourceID=ImageRanges[rangeID][1]-1;
			range,rangeID=GetResourceRange(resourceID);
		end
		return(resourceID);
	end

	self.FindUnknownNext=function(resourceID)
		-- get the resourcerange of the resourceID
		local range,rangeID=GetResourceRange(resourceID);
		while range>0 and resourceID<4294967295 do
			resourceID=ImageRanges[rangeID][2]+1;
			range,rangeID=GetResourceRange(resourceID);
		end
		return(resourceID);
	end

	self.imageMenu=Turbine.UI.ContextMenu();
	self.imageMenu:GetItems():Add(Turbine.UI.MenuItem("Category"))
	self.imageMenu:GetItems():Add(Turbine.UI.MenuItem("SubCategory"))
	self.imageMenu:GetItems():Add(Turbine.UI.MenuItem("Description"))
	self.imageMenu:GetItems():Get(1):SetEnabled(false);
	self.imageMenu:GetItems():Get(2):SetEnabled(false);
	self.imageMenu:GetItems():Get(3):SetEnabled(false);
	self.imageMenu:GetItems():Add(Turbine.UI.MenuItem("Edit Library Entry"));

	self.rangeMenu=Turbine.UI.ContextMenu();
	self.rangeMenu:GetItems():Add(Turbine.UI.MenuItem("Flag as Bad"))
	self.rangeMenu:GetItems():Add(Turbine.UI.MenuItem("Flag as Unknown"))
	self.rangeMenu:GetItems():Add(Turbine.UI.MenuItem("Flag as Good"))
	self.rangeMenu:GetItems():Get(1):SetEnabled(false);
	self.rangeMenu:GetItems():Get(2):SetEnabled(false);
	self.rangeMenu:GetItems():Get(3):SetEnabled(false);
	self.rangeMenu:GetItems():Get(1).Click=function()
		UpdateRange(tonumber(self.resourceID:GetText()),2,"C")
		irvWindow:ShowBackgrounds(irvWindow,false);
	end
	self.rangeMenu:GetItems():Get(2).Click=function()
		UpdateRange(tonumber(self.resourceID:GetText()),0,"C")
		irvWindow:ShowBackgrounds(irvWindow,false);
	end
	self.rangeMenu:GetItems():Get(3).Click=function()
		UpdateRange(tonumber(self.resourceID:GetText()),1,"C")
		irvWindow:ShowBackgrounds(irvWindow,false);
	end
	self.DisableControls=function()
		self.resourceID:SetEnabled(false);
		self.btnPriorRange:SetEnabled(false);
		self.btnPrevIDs:SetEnabled(false);
		self.btnGoToImage:SetEnabled(false);
		self.btnNextIDs:SetEnabled(false);
		self.btnNextRange:SetEnabled(false);
		self.btnForceView:SetEnabled(false);
		self.btnGridToggle:SetEnabled(false);
		-- we let the format toggle remain enabled since it's harmless
		self.btnSearch:SetEnabled(false);
		self.btnAutoPrev:SetEnabled(false);
		self.btnAutoNext:SetEnabled(false);
		self.btnIO:SetEnabled(false);

		self.imageRangeIndicator:SetMouseVisible(false); -- prevent popup menu
		self.imageSolo:SetMouseVisible(false);
	end
	self.EnableControls=function()
		self.resourceID:SetEnabled(true);
		self.btnPriorRange:SetEnabled(true);
		self.btnPrevIDs:SetEnabled(true);
		self.btnGoToImage:SetEnabled(true);
		self.btnNextIDs:SetEnabled(true);
		self.btnNextRange:SetEnabled(true);
		self.btnForceView:SetEnabled(true);
		self.btnGridToggle:SetEnabled(true);
		-- we let the format toggle remain enabled since it's harmless
		self.btnSearch:SetEnabled(true);
		self.btnAutoPrev:SetEnabled(true);
		self.btnAutoNext:SetEnabled(true);
		self.btnIO:SetEnabled(true);

		self.imageRangeIndicator:SetMouseVisible(true); -- prevent popup menu
		self.imageSolo:SetMouseVisible(true);
	end
end

function GetResourceRange(resource)
	-- determine which range this resource belongs in, Unknown=0, Safe=1, Danger=2
	local range=0;
	local index=1;
	local found=false;
	local passed=false
	while index<=#ImageRanges and not found and not passed do
		if resource>=ImageRanges[index][1] and resource<=ImageRanges[index][2] then
			if ImageRanges[index][3]=="S" then
				-- check for "SafeBoundaries"
				if UseSafeBoundaries then
					if resource~=ImageRanges[index][2] or (ImageRanges[index][4]~=nil and ImageRanges[index][4]~="D") then
						range=1;
					end
				else
					range=1;
				end
			else
				range=2;
			end
			found=true;
		else
			if resource<ImageRanges[index][1] then
				passed=true;
			else
				index=index+1;
			end
		end
	end
	return range,index
end
function UpdateRange(resource,class,type)
	local range,rangeID=GetResourceRange(resource);
	local priorRange,priorRangeID=GetResourceRange(resource-1);
	local nextRange,nextRangeID=GetResourceRange(resource+1);
	if type==nil then
		type="C";
	end

	if class~=range then
		if range==0 then
			-- see if there's a range below to attach to
			if priorRange==class then
				if nextRange==class then
					-- merge the ranges
					ImageRanges[priorRangeID][2]=ImageRanges[nextRangeID][2];
					table.remove(ImageRanges,nextRangeID);
				else
					-- just extend the prior range
					ImageRanges[priorRangeID][2]=resource;
				end
				if type=="C" then
					ImageRanges[priorRangeID][4]="C"; -- flag the range as Custom
				elseif type=="I" and ImageRanges[priorRangeID][4]=="D" then
					ImageRanges[priorRangeID][4]="I"; -- flag the range as imported only if it was Default previously
				end
			else
				if nextRange==class then
					-- just extend the next range
					ImageRanges[nextRangeID][1]=resource;
					if type=="C" then
						ImageRanges[nextRangeID][4]="C"; -- flag the range as Custom
					elseif type=="I" and ImageRanges[nextRangeID][4]=="D" then
						ImageRanges[nextRangeID][4]="I"; -- flag the range as imported only if it was Default previously
					end
				else
					-- no prior or next, need a new range
					-- this is an odd one since we need to know the right index to insert
					-- rangeid should point to one past the number of ranges or to the first range that is greater than our resourceid
					if class==1 then
						table.insert(ImageRanges,rangeID,{resource,resource,"S",type})
					elseif class==2 then
						table.insert(ImageRanges,rangeID,{resource,resource,"D",type})
					end
				end
			end
		else
			if class==0 then
				-- setting value back to "unknown", similar to swapping classes, but just a tiny bit different since we can't just slide it to a previous/next range
				if resource==ImageRanges[rangeID][1] then
					-- first entry
					if resource==ImageRanges[rangeID][2] then
						-- only entry, remove range
						table.remove(ImageRanges,rangeID);
					else
						-- just move low end up
						ImageRanges[rangeID][1]=resource+1;
						ImageRanges[rangeID][4]="C";
					end
				elseif resource==ImageRanges[rangeID][2] then
					-- can't be only entry since that would have been caught above, so just move hi end down
					ImageRanges[rangeID][2]=resource-1;
					ImageRanges[rangeID][4]="C";
				else
					-- isn't first or last, so we need to split the range creating a blank space
					table.insert(ImageRanges,rangeID+1,{})
					ImageRanges[rangeID+1][1]=resource+1;
					ImageRanges[rangeID+1][2]=ImageRanges[rangeID][2];
					ImageRanges[rangeID+1][3]=ImageRanges[rangeID][3];
					ImageRanges[rangeID+1][4]="C";
					ImageRanges[rangeID][2]=resource-1;
					ImageRanges[rangeID][4]="C";
				end
			else
				if resource==ImageRanges[rangeID][1] then
					if priorRange~=0 then
						ImageRanges[priorRangeID][2]=resource;
						ImageRanges[priorRangeID][4]="C";
						if ImageRanges[rangeID][2]==resource then
							table.remove(ImageRanges,rangeID);
							-- check for merge
							if nextRange~=0 then
								-- we already removed a range, so "next" range is now current range
								-- merge the ranges
								ImageRanges[priorRangeID][2]=ImageRanges[rangeID][2];
								table.remove(ImageRanges,rangeID);
							end
						else
							ImageRanges[rangeID][1]=resource+1;
							ImageRanges[rangeID][4]="C";
						end
					else
						if ImageRanges[rangeID][2]==resource then
							if class==1 then
								ImageRanges[rangeID][3]="S";
							elseif class==2 then
								ImageRanges[rangeID][3]="D";
							end
							ImageRanges[rangeID][4]="C";
						else
							ImageRanges[rangeID][1]=resource+1;
							if class==1 then
								table.insert(ImageRanges,rangeID,{resource,resource,"S","C"})
							elseif class==2 then
								table.insert(ImageRanges,rangeID,{resource,resource,"D","C"})
							end
						end
					end
				else

					if resource==ImageRanges[rangeID][2] then
						if nextRange~=0 then
							ImageRanges[nextRangeID][1]=resource;
							if ImageRanges[rangeID][1]==resource then
								table.remove(ImageRanges,rangeID);
							else
								ImageRanges[rangeID][2]=resource-1;
								ImageRanges[rangeID][4]="C";
							end
							ImageRanges[nextRangeID][4]="C";
						else
							if ImageRanges[rangeID][1]==resource then
								if class==1 then
									ImageRanges[rangeID][3]="S";
								elseif class==2 then
									ImageRanges[rangeID][3]="D";
								end
								ImageRanges[rangeID][4]="C";
							else
								ImageRanges[rangeID][2]=resource-1;
								ImageRanges[rangeID][4]="C";
								if class==1 then
									table.insert(ImageRanges,rangeID+1,{resource,resource,"S","C"})
								elseif class==2 then
									table.insert(ImageRanges,rangeID+1,{resource,resource,"D","C"})
								end
							end
						end
					else
						-- need to split our old range
						table.insert(ImageRanges,rangeID+1,{})
						ImageRanges[rangeID+1][1]=resource+1;
						ImageRanges[rangeID+1][2]=ImageRanges[rangeID][2];
						ImageRanges[rangeID+1][3]=ImageRanges[rangeID][3];
						ImageRanges[rangeID][2]=resource-1;
						ImageRanges[rangeID+1][4]="C";
						ImageRanges[rangeID][4]="C";
						if class==1 then
							table.insert(ImageRanges,rangeID+1,{resource,resource,"S","C"})
						elseif class==2 then
							table.insert(ImageRanges,rangeID+1,{resource,resource,"D","C"})
						end
					end
				end
			end
		end
	else
		if ImageRanges[rangeID][4]=="D" then
			ImageRanges[rangeID][4]=type;
		end
	end
end

function Window:ShowBackgrounds(sender,showUnknown)
	local resource;
	local range;
	self.btnForceView:SetVisible(false);
	if self.ShowGrid==true then
		local iRow,iCol;
		local imageWidth,imageHeight;
		local cellMaxWidth=self.imageGrid[1][1]:GetWidth()-12;
		local cellMaxHeight=self.imageGrid[1][1]:GetHeight()-30;
		self.imageSolo:SetVisible(false);
		self.imageRangeIndicator:SetVisible(false);
		self.imageSize:SetVisible(false);
		self.resourceDescription:SetText("");
		for iRow=1, #self.imageGrid do
			for iCol=1,#self.imageGrid[iRow] do
				self.imageGrid[iRow][iCol]:SetVisible(true);
				resource=tonumber(self.resourceID:GetText())+(iRow-1)*#self.imageGrid[1]+(iCol-1) ;
				self.imageGrid[iRow][iCol].imageID:SetText(string.format(self.numberFormat,resource));
				self.imageGrid[iRow][iCol].ResourceID=resource;
				range=GetResourceRange(resource);
				self.imageGrid[iRow][iCol].Status=range;
				if range==2 then
					self.imageGrid[iRow][iCol].imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/redlight.jpg");
					self.imageGrid[iRow][iCol].image:SetVisible(false);
					self.imageGrid[iRow][iCol].imageSize:SetText("");
-- display message that resource is in a range flagged as invalid
				else
					if range==1 then
						self.imageGrid[iRow][iCol].imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/greenlight.jpg");
						self.imageGrid[iRow][iCol].image:SetVisible(true);
						self.imageGrid[iRow][iCol].image:SetSize(1,1);
						self.imageGrid[iRow][iCol].image:SetBackground(resource);
						self.imageGrid[iRow][iCol].image:SetStretchMode(2); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
						imageWidth,imageHeight=self.imageGrid[iRow][iCol].image:GetSize();
						self.imageGrid[iRow][iCol].imageX=imageWidth;
						self.imageGrid[iRow][iCol].imageY=imageHeight;
						self.imageGrid[iRow][iCol].imageSize:SetText(imageWidth.." x "..imageHeight);
						self.imageGrid[iRow][iCol].image:SetStretchMode(0); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
						if imageWidth>cellMaxWidth then imageWidth=cellMaxWidth end;
						if imageHeight>cellMaxHeight then imageHeight=cellMaxHeight end;
						self.imageGrid[iRow][iCol].image:SetSize(imageWidth,imageHeight);
						self.imageGrid[iRow][iCol].image:SetPosition((cellMaxWidth-imageWidth)/2+1,(cellMaxHeight-imageHeight)/2+1);
					else
						self.imageGrid[iRow][iCol].imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/yellowlight.jpg");
						-- we ignore showUnknown in grid mode
						self.imageGrid[iRow][iCol].image:SetVisible(false);
						self.imageGrid[iRow][iCol].imageSize:SetText("");
-- display message that resource is in a range flagged as unkknown
					end
				end
			end
		end
	else
		for iRow=1, #self.imageGrid do
			for iCol=1,#self.imageGrid[iRow] do
				self.imageGrid[iRow][iCol]:SetVisible(false);
			end
		end
		self.imageRangeIndicator:SetVisible(true);
		self.imageSize:SetVisible(true);
		resource=tonumber(self.resourceID:GetText());
		self.resourceID:SetText(string.format(self.numberFormat,resource)); -- refresh the number using the current format
		range=GetResourceRange(resource);
		self.imageRangeIndicator.Range=range;
		self.imageSolo.Status=range;
		if range==2 then
			self.imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/redlight.jpg");
			self.imageSolo:SetVisible(false);
			self.imageSize:SetText("");
			self.resourceDescription:SetText("");
-- display message that resource is in a range flagged as invalid
		else
			if range==1 then
				self.imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/greenlight.jpg");
				self.imageSolo:SetVisible(true);
				self.imageSolo:SetSize(100,100);
				self.imageSolo:SetBackground(resource)
				self.imageSolo:SetStretchMode(2); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
				self.imageWidth,self.imageHeight=self.imageSolo:GetSize();
				self.imageSize:SetText(self.imageWidth.." x "..self.imageHeight);
				self.imageSolo:SetStretchMode(0); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
				self.imageSolo:SetSize(self.imageWidth,self.imageHeight); -- for some reason, we have to manually set the image size even though the width and height are reported as having changed already
				self.imageSolo:SetPosition((self:GetWidth()-self.imageWidth)/2,(self:GetHeight()-self.imageHeight)/2)
				local cat,subcat,desc=GetLibraryInfo(resource);
				if cat~="" then
					if subcat=="" then
						desc=cat.." - "..desc;
					else
						desc=cat.." / "..subcat.." - "..desc;
					end
				end

				self.resourceDescription:SetText(desc);
			else
				self.btnForceView:SetVisible(true);
				self.imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/yellowlight.jpg");
				self.resourceDescription:SetText("");
				if showUnknown==true then
					-- save any changes incase we blow up
					-- save the resource number as "bad resource" so we can automatically lock it out on reload
					irvWindow.TestID=resource;
					irvWindow:DisableControls(); -- disable the main window controls so the user doesn't affect the test
					-- display the message indicating that we are waiting to save our state before loading image
					self.UnknownImageWarning:SetVisible(true);
					SaveData(true);
					self.delayTime=Turbine.Engine:GetGameTime()+self.delayTotal;
					self.displayingWarning=true;
					self.UnknownImageWarning.status:SetWidth(0);
					self:SetWantsUpdates(true);  -- for some reason, the save function fires the dataSaveComplete event long before the data is actully saved :(
					self.btnForceView:SetVisible(false);
					-- the SaveComplete event handler will finish displaying the image once the settings are actually saved
				else
					self.imageSolo:SetVisible(false);
					self.imageSize:SetText("");
				end
			end
		end
	end
end
function SaveComplete()
	-- theoretically this handler is called when the data has been completely saved
	-- however, experience shows us that the event fires before the data is really saved to disk, so a client crash will still prevent the data from saving :(
	-- which renders this mechanism pretty much useless
end
function SaveComplete2()
local resource=irvWindow.TestID;
	if irvWindow:IsVisible() then
		-- function called when we completed saving data while testing a new image
		irvWindow.UnknownImageWarning:SetVisible(false);
		irvWindow.imageSolo:SetVisible(true);
		irvWindow.imageSolo:SetSize(100,100);
		irvWindow.imageSolo:SetBackground(resource)
		irvWindow.imageSolo:SetStretchMode(2); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
		irvWindow.imageWidth,irvWindow.imageHeight=irvWindow.imageSolo:GetSize();
		irvWindow.imageSize:SetText(irvWindow.imageWidth.." x "..irvWindow.imageHeight);
		irvWindow.imageSolo:SetStretchMode(0); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
		irvWindow.imageSolo:SetSize(irvWindow.imageWidth,irvWindow.imageHeight); -- for some reason, we have to manually set the image size even though the width and height are reported as having changed already
		irvWindow.imageSolo:SetPosition((irvWindow:GetWidth()-irvWindow.imageWidth)/2,(irvWindow:GetHeight()-irvWindow.imageHeight)/2)
		-- update the resource number as OK
		irvWindow.TestID=nil;
		Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_TestID",nil);
		UpdateRange(resource,1);
		irvWindow.imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/greenlight.jpg");

		-- implement the AutoPrev and AutoNext decrement/increment here
		if irvWindow.AutoPrev==true then
			irvWindow.resourceID:SetText(string.format(irvWindow.numberFormat,irvWindow.FindUnknownPrev(resource)));
			if tonumber(irvWindow.resourceID:GetText())>0 then
				irvWindow.AutoNext=false; -- just in case...
				-- get prev image. the rest of the 'auto' code is in SaveComplete2()
				irvWindow:ShowBackgrounds(irvWindow,true);
			end
		elseif irvWindow.AutoNext==true then
			irvWindow.resourceID:SetText(string.format(irvWindow.numberFormat,irvWindow.FindUnknownNext(resource)));
			if tonumber(irvWindow.resourceID:GetText())>0 then
				irvWindow.AutoPrev=false; -- just in case...
				-- get next image.
				irvWindow:ShowBackgrounds(irvWindow,true);
			end
		else
			irvWindow:EnableControls();
		end
	else
		-- we can't allow the test to complete if the window is not visible since the failure only occurs when the window is visible
		irvWindow.UnknownImageWarning:SetVisible(false);
		irvWindow.TestID=nil;
		Turbine.PluginData.Save( Turbine.DataScope.Account, "IRV_TestID",nil);
		irvWindow.AutoPrev=false;
		irvWindow.AutoNext=false;		
		irvWindow:EnableControls();
	end
end

function TestComplete()
	-- routine for completing the SafeBoundaries test
local resource=0x41007ee4; -- always test with this one "known" problem boundary
	if irvWindow:IsVisible() then
		irvWindow.UnknownImageWarning:SetVisible(false);
		irvWindow.imageSolo:SetVisible(true);
		irvWindow.imageSolo:SetSize(100,100);
		-- attempt to assign the resource as a background... if we're gonna blow up it will be here
		irvWindow.imageSolo:SetBackground(resource)
		irvWindow.imageSolo:SetStretchMode(2); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
		irvWindow.imageWidth,irvWindow.imageHeight=irvWindow.imageSolo:GetSize();
		irvWindow.imageSize:SetText(irvWindow.imageWidth.." x "..irvWindow.imageHeight);
		irvWindow.imageSolo:SetStretchMode(0); -- setting stretchmode(2) AFTER loading an image will resize the control to the size of the image
		irvWindow.imageSolo:SetSize(irvWindow.imageWidth,irvWindow.imageHeight); -- for some reason, we have to manually set the image size even though the width and height are reported as having changed already
		irvWindow.imageSolo:SetPosition((irvWindow:GetWidth()-irvWindow.imageWidth)/2,(irvWindow:GetHeight()-irvWindow.imageHeight)/2)
		-- turn off safeboundaries
		UseSafeBoundaries=false;
		SaveData();
		UpdateRange(resource,1);
		irvWindow.imageRangeIndicator:SetBackground("GaranStuff/IRV/Resources/greenlight.jpg");
		irvWindow:EnableControls();
		irvWindow.UnknownImageWarning:SetText("You are attempting to display a previously unverified image.\n\nThe system is currently saving all data in case the image causes a client crash.  As soon as the current data is saved, the image will be displayed.\n\nIf the client does crash, this image will automatically be locked out from displaying in the future (to avoid additional crashes). If the image loads without incident, the image will be flagged as good for future displays.");
		irvWindow.UnknownImageWarning.btnSkip:SetVisible(true);
		irvWindow.UnknownImageWarning.btnAbort:SetVisible(true);
	end
end

function GetLibraryInfo(resourceID)
	local category="";
	local subcategory="";
	local description="Undefined";
	local found=false;
	local libIndex=1;
	-- find the resource and return it's Class, Subclass and Description
	-- we have to do a table scan since the table is indexed by Category/Description not resourceID :(
	while libIndex<=#Library and not found do
		if resourceID==Library[libIndex][1] then
			found=true;
		else
			libIndex=libIndex+1;
		end
	end
	if found then
		description=Library[libIndex][4];
		found=false;
		-- now look up the Category;
		local catIndex=1;
		local subCatIndex=1;
		while not found and catIndex<=#Categories do
			if Library[libIndex][5]==Categories[catIndex][1] then
				found=true;
			else
				catIndex=catIndex+1;
			end
		end
		if found then
			category=Categories[catIndex][2];
			if Library[libIndex][6]~=nil then
				found=false;
				local subCat=Categories[catIndex][3];
				while not found and subCatIndex<=#SubCategories[subCat] do
					if Library[libIndex][6]==SubCategories[subCat][subCatIndex][1] then
						found=true;
					else
						subCatIndex=subCatIndex+1;
					end
				end
				if found then
					subcategory=SubCategories[subCat][subCatIndex][2];
				end
			end
		end
	end
	return category,subcategory,description;
end
-- used to export the Update ImageRanges data
function DumpImageRangesToChat()
	local index=1
	while index<=#ImageRanges do
		if ImageRanges[index][4]=="C" then
			Turbine.Shell.WriteLine("IR[#IR+1]={"..string.format("0x%x",ImageRanges[index][1])..","..string.format("0x%x",ImageRanges[index][2])..", \""..ImageRanges[index][3].."\", \"D\"}" )
		end
		index=index+1
	end
end

-- used to export the Update Library data
function DumpLibraryToChat()
	local index=1
	while index<=#Library do
		if Library[index][7]=="C" then
			Turbine.Shell.WriteLine("UL[#UL+1]={"..string.format("0x%x",Library[index][1])..","..tostring(Library[index][2])..", "..tostring(Library[index][3])..", \""..tostring(Library[index][4]).."\", "..tostring(Library[index][5])..", "..tostring(Library[index][6]).."}" )
		end
		index=index+1
	end
end

LoadData(); -- we want to have all of the categories and other arrays filled before we try to instantiate the windows since some controls depend on this data (like the category list)

if #ImageRanges==0 or #Categories==0 or #Library==0 then
import "GaranStuff.IRV.Defaults";
LoadDefaults()
end

-- Instantiate the actual windows, etc

irvWindow=Window();
libWindow=LibraryWindow();
libWindow:SetVisible(false);
editLibWindow=EditLibraryWindow();
editLibWindow:SetVisible(false);
catMaintWindow=CategoryMaintWindow();
catMaintWindow:SetVisible(false);
ioWindow=IOWindow();
ioWindow:SetVisible(false);

irvWindow:SetZOrder(irvWindow:GetZOrder()+1); -- force irv to "Topmost"
libWindow:SetZOrder(irvWindow:GetZOrder()+1);
editLibWindow:SetZOrder(libWindow:GetZOrder()+1);
catMaintWindow:SetZOrder(editLibWindow:GetZOrder()+1);
ioWindow:SetZOrder(catMaintWindow:GetZOrder()+1);

--DumpImageRangesToChat(); -- used to generate the Updates ranges for builds
--DumpLibraryToChat(); -- used to generate the Updates ranges for builds

irvWindow:SetVisible(true);

irvWindow:ShowBackgrounds(false);

